import re
import sys
import threading
import collections
from httplib2 import Http
try:
    from urllib.parse import urlsplit, urljoin
except ImportError:
    # Python 2
    from urlparse import urlsplit, urljoin

import webob

__all__ = ['Policy', 'AkamaiPolicy', 'MiddleWare', 'InvalidESIMarkup', 'RecursionError']

try:
    from sys import getsizeof
except ImportError:
    # Python 2.5
    def getsizeof(obj):
        if isinstance(obj, basestring):
            # approximation for strings, which is what httplib stores
            return len(obj)
        return 0

try:
    basestring
except NameError:
    basestring = str

#
# Policies that can make the middleware work like different ESI processors
#

class Policy(object):
    max_nested_includes = None
    chase_redirect = False
    cache = None

    def http(self):
        http = Http(cache=self.cache, timeout=5, disable_ssl_certificate_validation=True)
        http.follow_redirects = self.chase_redirect
        return http

class AkamaiPolicy(Policy):
    """Configure the middleware to behave like akamai"""
    max_nested_includes = 5



#: Client headers to forward in subrequests to all servers
forward_headers_all_servers = set(['accept-language', 'cache-control'])

#: Client headers to forward only in subrequests to the same server
forward_headers_same_origin = forward_headers_all_servers | \
        set(['cookie', 'authorization', 'referer'])

#
# Cache
#

_marker = object()

class _Counter(dict):

    def __missing__(self, key):
        return 0

class LRUCache(object):

    def __init__(self, maxsize=1000, max_object_size=102400):
        # 1000 * 40kb/page ~ 40Mb
        maxqueue = maxsize * 10
        queuedrop = maxsize * 2
        # set instance variables so we can test
        self._cache = cache = {}
        self._refcount = refcount = _Counter()
        self._queue = queue = collections.deque()
        lock = threading.Lock()
        self.hits = 0
        self.misses = 0

        def compact_queue():
            # compact the queue when it gets too big
            # first: remove duplicates
            refcount.clear()
            queue.appendleft(_marker)
            for k in iter(queue.pop, _marker):
                if k in refcount:
                    continue
                queue.appendleft(k)
                refcount[k] = 1
            if len(queue) > maxqueue:
                # if we're still too big, and have no duplicates
                # there's probably something hammering the same thing remove
                # queuedrop items not in our cache
                count = 0
                queue.append(_marker)
                while count <= queuedrop:
                    key = queue.popleft()
                    assert key is not _marker
                    if key in self._cache:
                        queue.append(key)
                    else:
                        count += 1
                        del refcount[key]
                for k in iter(queue.popleft, _marker):
                    queue.append(k)

        def get(key):
            if lock.acquire(False):
                try:
                    queue.append(key)
                    refcount[key] = refcount.get(key, 0) + 1
                    if len(queue) > maxqueue:
                        compact_queue()
                finally:
                    lock.release()
            val = cache.get(key, _marker)
            if val is not _marker:
                self.hits += 1
                return val
            self.misses += 1
            return None

        def set(key, value):
            if max_object_size is not None and getsizeof(value) > max_object_size:
                # note, this doesn't take into account the size of objects referenced by value
                return
            orig_key = key
            if len(cache) >= maxsize:
                # remove least recently used
                key = queue.popleft()
                refcount[key] -= 1
                while refcount[key]:
                    key = queue.popleft()
                    refcount[key] -= 1
                del refcount[key]
                delete(key)
            queue.appendleft(orig_key)
            refcount[orig_key] += 1
            cache[orig_key] = value

        def locked_set(key, value):
            lock.acquire()
            try:
                set(key, value)
            finally:
                lock.release()

        def delete(key):
            cache.pop(key, None)

        self.get = get
        self.set = locked_set
        self.delete = delete

#
# The middleware
#

class MiddleWare(object):

    def __init__(self, app, policy='default', debug=True):
        self.debug = debug
        self.app = app
        if isinstance(policy, basestring):
            policy = _POLICIES[policy]
        self.policy = policy
        self.http = policy.http()

    def __call__(self, environ, start_response):
        req = webob.Request(environ)
        resp = req.get_response(self.app)
        if resp.content_type == 'text/html' and resp.status_int == 200:
            new_body = self._process(resp.body, req)
            if new_body is not None:
                resp.body = new_body
        return resp(environ, start_response)

    def _process(self, body, req):
        commented = self._commented(body)
        return self._process_include(body, req, comments=commented)

    def _commented(self, body):
        # identify parts of body which are comments
        comments = []
        c_idx = 0

        # Compatibility workaround: in python 2 this returns ``'>'``,
        # in python 3 it's ``62``.
        end_of_comment_marker = b'>'[0]
        while 1:
            match = _re_comment.search(body, c_idx)
            if match is None:
                break
            c_idx = match.start() + 1
            if len(body) < match.end() + 1:
                continue
            if body[match.end()] != end_of_comment_marker:
                # invalid comment, contains --, ignore it
                continue
            # we found a comment
            c_idx = match.end()
            comments.append((match.start(), match.end() + 1))
        return tuple(comments)

    def _process_include(self, body, req, level=0, comments=()):
        debug = self.debug
        policy = self.policy
        comments = list(comments)
        require_ssl = not (req.environ['wsgi.url_scheme'] == 'http')
        if debug and policy.max_nested_includes is not None and level > policy.max_nested_includes:
            raise RecursionError('Too many nested includes', level, body)
        c_start = c_end = None
        if comments:
            c_start, c_end = comments.pop(0)
        # process the includes
        index = 0
        new = []
        matches = _re_include.finditer(body)
        for match in matches:
            if c_end is not None:
                while c_end is not None and c_end < match.end():
                    # remove comments which we have passed
                    c_start = c_end = None
                    if comments:
                        c_start, c_end = comments.pop(0)
                if c_end is not None:
                    # ignore this match if we are in a comment
                    if c_start < match.start() and c_end > match.end():
                        continue
            # add section before current match to new body
            new.append(body[index:match.start()])
            if match.group('other') or not match.group('src'):
                if debug:
                    raise InvalidESIMarkup("Invalid ESI markup: %s" % body[match.start():match.end()])
                # silently ignore this match
                index = match.end()
                continue
            # get content to insert
            try:
                new_content = _include_url(match.group('src'), req, require_ssl, policy.chase_redirect, self.http)
            except:
                if match.group('alt'):
                    try:
                        new_content = _include_url(match.group('alt'), req, require_ssl, policy.chase_redirect, self.http)
                    except:
                        if match.group('onerror') == b'continue':
                            new_content = b''
                        else:
                            raise
                elif match.group('onerror') == b'continue':
                    new_content = b''
                else:
                    raise
            if new_content:
                # recurse to process any includes in the new content
                new_commented = self._commented(new_content)
                p = self._process_include(new_content, req, comments=new_commented, level=level + 1)
                if p is not None:
                    new_content = p
            new.append(new_content)
            # update index
            index = match.end()
        if not index:
            return None
        new.append(body[index:])
        return b''.join(new)

#
# Exceptions we can raise
#

class InvalidESIMarkup(Exception):
    pass

class RecursionError(Exception):

    def __init__(self, msg, level, body):
        super(RecursionError, self).__init__(msg, level, body)
        self.msg = msg
        self.body = body
        self.level = level

class IncludeError(Exception):
    pass

#
# The internal bits to do the work
#

_POLICIES = {'default': Policy(),
             'akamai': AkamaiPolicy()}

_re_include = re.compile(br'''<esi:include'''
                         br'''(?:\s+(?:''' # whitespace at start of tag
                             br'''src=["']?(?P<src>[^"'\s]*)["']?''' # find src=
                             br'''|alt=["']?(?P<alt>[^"'\s]*)["']?''' # or find alt=
                             br'''|onerror=["']?(?P<onerror>[^"'\s]*)["']?''' # or find onerror=
                             br'''|(?P<other>[^\s><]+)?''' # or find something eles
                         br'''))+\s*/>''') # match whitespace at the end and the end tag

_re_comment = re.compile(br'''<!--esi.*?--''', flags=re.DOTALL)

class _HTTPError(Exception):

    def __init__(self, url, status):
        self.status = status
        message = 'Url returned %s: %s' % (status, url)
        super(_HTTPError, self).__init__(message)


def _forward_all_headers_allowed(origin_host, is_ssl, url):
    """
    Return True if headers can be forwarded to urlparse result ``url``.
    This returns true if ``url`` refers to the same server and same protocol
    (http, https) as ``origin_host``.

    This is overly simplistic,
    """

    # Fail safe in the case that the original host header was not specified
    if not origin_host:
        return False

    # Don't allow headers to be fowarded from https -> http or vice versa
    if is_ssl != bool(url.scheme == 'https'):
        return False

    url_host = url.netloc

    if ':' not in url_host:
        url_host += ':443' if is_ssl else ':80'

    if ':' not in origin_host:
        origin_host += ':443' if is_ssl else ':80'

    return url_host == origin_host


def _include_url(orig_url, req, require_ssl, chase_redirect, http):
    orig_url = orig_url.decode('ascii')
    orig_url = urljoin(req.path_url, orig_url)
    url = urlsplit(orig_url)
    headers = req.headers
    if require_ssl and url.scheme != 'https':
        raise IncludeError('SSL required, cannot include: %s' % (orig_url, ))

    if _forward_all_headers_allowed(headers.get('Host'), require_ssl, url):
        forward_headers = forward_headers_same_origin
    else:
        forward_headers = forward_headers_all_servers

    headers = dict((k, v)
                    for k, v in headers.items()
                    if k.lower() in forward_headers)

    resp, content = http.request(orig_url, headers=dict(headers))
    if resp.status == 200:
        return content
    raise _HTTPError(orig_url, resp.status)



import os
from unittest import TestCase

import webob
from mock import patch, Mock, call

all_tests = False
if os.environ.get('WESGI_ALL_TESTS', 'false').lower() in ('true', '1', 't'):
    all_tests = True

patch_Http = patch('wesgi.Http', mocksignature=True)

def mock_http_request(http, response=None, content=None):
    """Return a httplib2.Http with request() mocked out"""
    http.request = Mock(spec_set=[])
    mock = http.request
    if content is not None or response is not None:
        if content is None:
            content = b''
        if response is None:
            response = Response()
        mock.return_value = response, content

def Response(status=200, headers=None):
    import httplib2
    assert isinstance(status, int)
    d = dict(status=str(status))
    if headers is not None:
        d.update(headers)
    return httplib2.Response(d)

def run_mw(mw, **requestkwargs):
    start_response = Mock()
    request = webob.Request.blank("", **requestkwargs)
    response = mw(request.environ, start_response)
    return b''.join(response)

def make_mw(app=None, **kw):
    status = kw.pop('http_status', 200)
    headers = kw.pop('http_headers', None)
    content = kw.pop('http_content', b'')
    if app is None:
        body = kw.pop('app_body', b'')
        app = make_app(body=body)
    from wesgi import MiddleWare
    mw = MiddleWare(app, **kw)
    mock_http_request(mw.http, Response(status=status, headers=headers), content)
    return mw

def make_app(body=b'', content_type='text/html', status=200):
    def _app(environ, start_response):
        response = webob.Response(body, content_type=content_type)
        response.status = status
        return response(environ, start_response)
    return _app


class TestProcessInclude(TestCase):

    def test_return_none_if_no_match(self):
        mw = make_mw()
        req = webob.Request.blank("")
        mock = mw.http.request
        data = mw._process_include(b'', req)
        self.assertEqual(data, None)
        self.assertFalse(mock.called)
        data = mw._process_include(b'something', req)
        self.assertEqual(data, None)
        self.assertFalse(mock.called)
        data = mw._process_include(b'<html><head></head><body><h1>HI</h1><esi:not_an_include whatever="bobo"/></body></html>', req)
        self.assertEqual(data, None)
        self.assertFalse(mock.called)

    def test_match(self):
        mw = make_mw(http_content=b'<div>example</div>')
        req = webob.Request.blank("")
        data = mw._process_include(b'before<esi:include src="http://www.example.com"/>after', req)
        self.assertEqual(data, b'before<div>example</div>after')
        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(mw.http.request.call_args,
                         call('http://www.example.com', headers={}))
        # onerror="continue" has no effect
        mw.http.request.reset_mock()
        data = mw._process_include(b'before<esi:include src="http://www.example.com" onerror="continue"/>after', req)
        self.assertEqual(data, b'before<div>example</div>after')
        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(mw.http.request.call_args,
                         call('http://www.example.com', headers={}))

    def test_recursive(self):
        from wesgi import RecursionError, AkamaiPolicy
        counter = list(range(10))
        def side_effect(*args, **kwargs):
            if not counter:
                return Response(), b'-last'
            count = counter.pop(0)
            fragment = count + 1 # the nth fragment that we are including
            count += 2 # count starts at 0, but first time we include is level 2
            return (Response(),
                    ('-%s<esi:include src="http://www.example.com/%s"/>' %
                     (count, fragment)).encode('ascii'))
        req = webob.Request.blank("")
        mw = make_mw()
        mock = mw.http.request
        mock.side_effect = side_effect
        data = mw._process_include(b'level-1<esi:include src="http://www.example.com"/>-after', req)
        self.assertEqual(data, b'level-1-2-3-4-5-6-7-8-9-10-11-last-after')
        self.assertEqual(mock.call_count, 11)
        self.assertEqual(mock.call_args,
                         call('http://www.example.com/10', headers={}))
        # Akamai FAQ http://www.akamai.com/dl/technical_publications/esi_faq.pdf
        # claims that they support 5 levels of nested includes. Ours should do the same
        # when using the akami policy
        counter.extend(range(5))
        mw.policy = AkamaiPolicy()
        self.assertRaises(RecursionError, mw._process_include, b'level-1<esi:include src="http://www.example.com"/>-after', req)
        counter.extend(range(4))
        data = mw._process_include(b'level-1<esi:include src="http://www.example.com"/>-after', req)
        self.assertEqual(data, b'level-1-2-3-4-5-last-after')
        self.assertEqual(mw.http.request.call_args,
                         call('http://www.example.com/4', headers={}))
        # Even with the akamai policy, if we're not in debug mode, no error is raised
        counter.extend(range(10))
        mw.debug = False
        data = mw._process_include(b'level-1<esi:include src="http://www.example.com"/>-after', req)
        self.assertEqual(data, b'level-1-2-3-4-5-6-7-8-9-10-11-last-after')

    def test_invalid(self):
        from wesgi import InvalidESIMarkup
        req = webob.Request.blank("")
        mw = make_mw()
        invalid1 = b'before<esi:include krud src="http://www.example.com"/>after'
        invalid2 = b'before<esi:include krud="krud" src="http://www.example.com"/>after'
        self.assertRaises(InvalidESIMarkup, mw._process_include, invalid1, req)
        self.assertRaises(InvalidESIMarkup, mw._process_include, invalid2, req)
        # if debug is False, these errors are not raised
        mw.debug = False
        self.assertEqual(mw._process_include(invalid1, req), b'beforeafter')
        self.assertEqual(mw._process_include(invalid2, req), b'beforeafter')
        self.assertFalse(mw.http.request.called)

    def test_some_http_error_cases(self):
        class Oops(Exception):
            pass
        def side_effect(*args, **kwargs):
            def second_call(*args, **kwargs):
                return Response(), b'<div>example alt</div>'
            mw.http.request.side_effect = second_call
            raise Oops('oops')
        req = webob.Request.blank("")
        mw = make_mw()
        mw.http.request.side_effect = side_effect
        # without src we get our exception
        self.assertRaises(Oops, mw._process_include, b'before<esi:include src="http://www.example.com"/>after', req)
        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(mw.http.request.call_args, call('http://www.example.com', headers={}))
        # it is still raised if we turn off debug mode (it's specified in the ESI spec)
        mw.http.request.side_effect = side_effect
        mw.debug = False
        self.assertRaises(Oops, mw._process_include, b'before<esi:include src="http://www.example.com"/>after', req)
        # unless onerror="continue", in which case the include is silently deleted
        mw = make_mw()
        mw.http.request.side_effect = side_effect
        data = mw._process_include(b'before<esi:include src="http://www.example.com" onerror="continue"/>after', req)
        self.assertEqual(data, b'beforeafter')
        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(mw.http.request.call_args, call('http://www.example.com', headers={}))
        # if we add a alt we get back the info from alt
        mw = make_mw()
        mw.http.request.side_effect = side_effect
        data = mw._process_include(b'before<esi:include src="http://www.example.com" alt="http://alt.example.com"/>after', req)
        self.assertEqual(data, b'before<div>example alt</div>after')
        self.assertEqual(mw.http.request.call_args_list,
                          [call('http://www.example.com', headers={}),
                           call('http://alt.example.com', headers={})])
        # onerror = "continue" has no effect if there is only one error and alt is specified
        mw = make_mw()
        mw.http.request.side_effect = side_effect
        data = mw._process_include(b'before<esi:include src="http://www.example.com" alt="http://alt.example.com" onerror="continue"/>after', req)
        self.assertEqual(data, b'before<div>example alt</div>after')
        self.assertEqual(mw.http.request.call_args_list,
                          [call('http://www.example.com', headers={}),
                           call('http://alt.example.com', headers={})])
        # If both calls to mw.http.request fail, the second exception is raised
        class OopsAlt(Exception):
            pass
        def side_effect(*args, **kwargs):
            def second_call(*args, **kwargs):
                raise OopsAlt('oops')
            mw.http.request.side_effect = second_call
            raise Oops('oops')
        mw = make_mw()
        mw.http.request.side_effect = side_effect
        self.assertRaises(OopsAlt, mw._process_include, b'before<esi:include src="http://www.example.com" alt="http://alt.example.com"/>after', req)
        self.assertEqual(mw.http.request.call_args_list,
                         [call('http://www.example.com', headers={}),
                          call('http://alt.example.com', headers={})])
        # it is still raised if we turn off debug mode (it's specified in the ESI spec)
        mw.http.request.side_effect = side_effect
        mw.debug = False
        self.assertRaises(OopsAlt, mw._process_include, b'before<esi:include src="http://www.example.com" alt="http://alt.example.com"/>after', req)
        # unless onerror="continue", in which case the include is silently deleted
        mw = make_mw()
        mw.http.request.side_effect = side_effect
        data = mw._process_include(b'before<esi:include src="http://www.example.com" alt="http://alt.example.com" onerror="continue"/>after', req)
        self.assertEqual(data, b'beforeafter')
        self.assertEqual(mw.http.request.call_args_list,
                         [call('http://www.example.com', headers={}),
                          call('http://alt.example.com', headers={})])

    def test_regression_regex_performance_extra_data(self):
        # processing this data used to take a LOONG time
        import time
        req = webob.Request.blank("")
        mw = make_mw(http_content=b'<div>example</div>')
        this_dir = os.path.dirname(__file__)
        test_data = b'<esi:include src="http://www.google.com" />\n\t\t\r\n\t\t\t\r\n\t\t\t\t\r\n\t\t\t\t\r\n\t\t\t\r\n\t\t</p>\r\n'
        now = time.time()
        data = mw._process_include(test_data, req)
        used = time.time() - now
        self.assertTrue(used < 0.01, 'Test took too long: %s seconds' % used)


class TestMiddleWare(TestCase):

    def test_process(self):
        mw = make_mw(app_body=b'before<esi:include src="http://www.example.com"/>after',
                     http_content=b"<div>example</div>")

        response = run_mw(mw)

        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(mw.http.request.call_args,
                          call('http://www.example.com', headers={}))
        self.assertEqual(response, b'before<div>example</div>after')

    def test_process_ssl(self):
        from wesgi import IncludeError
        mw = make_mw(app_body=b'before<esi:include src="http://www.example.com"/>after',
                     http_content=b'<div>example</div>')

        # trying to include an http: url from an https page raises an error
        start_response = Mock()
        request = webob.Request.blank("")
        request.environ['wsgi.url_scheme'] = 'https'
        response = self.assertRaises(IncludeError, mw, request.environ, start_response)
        self.assertEqual(mw.http.request.call_count, 0)

        # https urls do work
        mw = make_mw(app_body=b'before<esi:include src="https://www.example.com"/>after',
                     http_content=b'<div>example</div>')
        response = mw(request.environ, start_response)
        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(mw.http.request.call_args,
                          call('https://www.example.com', headers={}))
        self.assertEqual(b''.join(response), b'before<div>example</div>after')

    def test_comment(self):
        result = b'<div>example</div>'
        this_dir = os.path.dirname(__file__)
        include = b'<esi:include src="http://www.example.com" />'
        test_data = [(include, result),
                     b'<!-- html comment',
                     (include, result),
                     b'-->',
                     b'blah',
                     b'<!--esi half open esi',
                     (include, result),
                     b'<!--esi esi comment 1',
                     (include, include),
                     b'-->',
                     (include, result),
                     b'<!--esi esi comment 2',
                     (include, include),
                     b'-->',
                     b'<!--esi esi comment 3 containing a single -',
                     (include, include),
                     b'-->',
                     (include, result)]
        expected = []
        data = []
        for i in test_data:
            input = res = i
            if not isinstance(i, bytes):
                input, res = i
            expected.append(res)
            data.append(input)
        req = webob.Request.blank("")
        mw = make_mw(http_content=result,
                     app_body=b'\n'.join(data))
        data = run_mw(mw)
        expected = b'\n'.join(expected)
        self.assertEqual(data, expected)
        # regression test for an error
        self.assertEqual(mw._process_include(b'<!--esi ' + include + b' --', req),
                                              b'<!--esi ' + result + b' --')

    def test_it_forwards_request_headers(self):
        mw = make_mw(app_body=b'<esi:include src="http://www.example.com"/>')
        run_mw(mw, headers={'Host': 'www.example.com:80',
                            'Cookie': 'x'})
        self.assertEqual(mw.http.request.call_args,
                         call('http://www.example.com',
                              headers={'Cookie': 'x'}))

    def test_it_doesnt_forward_inappropriate_headers(self):
        """
        Relaying eg Content-Length from a POST request will cause the
        subrequest to timeout while the server waits for the non-existant
        content
        """
        mw = make_mw(app_body=b'<esi:include src="http://www.example.com"/>')
        run_mw(mw, headers={'Host': 'www.example.com',
                            'Content-Length': '100'})
        self.assertEqual(mw.http.request.call_args,
                         call('http://www.example.com', headers={}))

    def test_it_forwards_non_sensitive_headers_only_to_non_origin_server(self):
        """
        Relaying eg cookies from one server to another is not allowed
        """
        mw = make_mw(app_body=b'<esi:include src="http://www.example.net"/>')
        run_mw(mw, headers={'Host': 'www.example.com',
                            'Cache-Control': 'no-cache',
                            'Cookie': 'private'})
        self.assertEqual(mw.http.request.call_args,
                         call('http://www.example.net',
                              headers={'Cache-Control': 'no-cache'}))

    def test_relative_links_are_turned_into_absolute_for_backend_request(self):
        mw = make_mw(app_body=b'<esi:include src="/relative/url"/>')

        run_mw(mw, headers={'Host': 'www.example.com'})

        self.assertEqual(mw.http.request.call_count, 1)
        self.assertEqual(
                mw.http.request.call_args[0],
                ('http://www.example.com/relative/url', ))


class TestPolicy(TestCase):

    def test_chase_redirect(self):
        from wesgi import Policy
        policy = Policy()
        self.assertEqual(policy.http().follow_redirects, False)
        # unless it's specified in the policy
        policy = Policy()
        policy.chase_redirect = True
        self.assertEqual(policy.http().follow_redirects, True)

    def test_cache(self):
        # no caching by default
        from wesgi import Policy
        policy = Policy()
        self.assertEqual(policy.cache, None)

class TestLRUCache(TestCase):

    def test_basic(self):
        from wesgi import LRUCache
        cache = LRUCache()
        self.assertEqual(cache.get('a'), None)
        self.assertEqual(cache.get('b'), None)
        self.assertEqual(cache._refcount, {'a': 1, 'b': 1})
        self.assertEqual(cache._cache, {})
        cache.set('a', 'x')
        self.assertInvariants(cache)
        self.assertEqual(cache.get('a'), 'x')
        self.assertEqual(cache.get('b'), None)
        self.assertEqual(cache._refcount, {'a': 3, 'b': 2})
        self.assertEqual(cache._cache, {'a': 'x'})
        cache.set('b', 'y')
        self.assertInvariants(cache)
        self.assertEqual(cache.get('a'), 'x')
        self.assertEqual(cache.get('b'), 'y')
        self.assertEqual(cache._refcount, {'a': 4, 'b': 4})
        self.assertEqual(cache._cache, {'a': 'x', 'b': 'y'})
        cache.set('b', 'z')
        self.assertInvariants(cache)
        self.assertEqual(cache.get('a'), 'x')
        self.assertEqual(cache.get('b'), 'z')
        self.assertEqual(cache._refcount, {'a': 5, 'b': 6})
        self.assertEqual(cache._cache, {'a': 'x', 'b': 'z'})
        cache.delete('b')
        self.assertInvariants(cache)
        self.assertEqual(cache._refcount, {'a': 5, 'b': 6})
        self.assertEqual(cache.get('a'), 'x')
        self.assertEqual(cache.get('b'), None)
        self.assertEqual(cache._cache, {'a': 'x'})
        cache.delete('a')
        self.assertInvariants(cache)
        self.assertEqual(cache._refcount, {'a': 6, 'b': 7})
        self.assertEqual(cache.get('a'), None)
        self.assertEqual(cache.get('b'), None)
        self.assertEqual(cache._cache, {})

    def test_max_object_size(self):
        from wesgi import LRUCache
        cache = LRUCache(maxsize=3, max_object_size=1000)
        # objects larger than max_object_size are ignored
        cache.set('a', 'a' * 2000)
        self.assertEqual(cache._cache, {})
        # smaller objects are not
        cache.set('a', 'a')
        self.assertEqual(cache._cache, {'a': 'a'})

    def test_hit_miss(self):
        # an LRU's biggest weakness is the sequential scan
        # this is what happens
        from wesgi import LRUCache
        cache = LRUCache(maxsize=3)
        cache.get('a')
        cache.set('a', 'a')
        self.assertEqual(cache.hits, 0)
        self.assertEqual(cache.misses, 1)
        cache.get('a')
        self.assertEqual(cache.hits, 1)
        self.assertEqual(cache.misses, 1)
        cache.get('b')
        self.assertEqual(cache.hits, 1)
        self.assertEqual(cache.misses, 2)
        cache.get('a')
        self.assertEqual(cache.hits, 2)
        self.assertEqual(cache.misses, 2)
        self.assertInvariants(cache)

    def test_repeated_get_and_set_flushes_cache(self):
        # an LRU's biggest weakness is the sequential scan
        # this is what happens
        from wesgi import LRUCache
        cache = LRUCache(maxsize=3)
        cache.get('a')
        cache.set('a', 'a')
        cache.get('b')
        cache.set('b', 'b')
        for i in range(100):
            cache.get(str(i))
            cache.set(str(i), str(i))
        self.assertEqual(cache._cache, {'99': '99', '98': '98', '97': '97'})
        self.assertEqual(cache.hits, 0)
        self.assertEqual(cache.misses, 102)
        self.assertInvariants(cache)

    def test_repeated_set_without_get_does_not_flushe_cache(self):
        from wesgi import LRUCache
        cache = LRUCache(maxsize=3)
        cache.get('a')
        cache.set('a', 'a')
        cache.get('b')
        cache.set('b', 'b')
        for i in range(100):
            cache.set(str(i), str(i))
        self.assertEqual(cache._cache, {'99': '99', 'a': 'a', 'b': 'b'})
        self.assertInvariants(cache)

    def test_queue_compaction(self):
        from wesgi import LRUCache
        cache = LRUCache(maxsize=10)
        [cache.get(1) for i in range(100)] # fill up to maxqueue
        self.assertEqual(list(cache._queue), [1 for i in range(100)]) # our queue is full
        self.assertEqual(cache._refcount, {1: 100})
        cache.get(1) #push us over the limit
        self.assertEqual(list(cache._queue), [1])
        self.assertEqual(cache._refcount, {1: 1})
        self.assertInvariants(cache)

    def test_queue_comaction_different_values(self):
        from wesgi import LRUCache
        # test compaction with a different value first
        cache = LRUCache(maxsize=10)
        cache.get(2)
        [cache.get(1) for i in range(49)]
        cache.get(2)
        [cache.get(1) for i in range(50)] # Push us one over the limit
        self.assertEqual(list(cache._queue), [2, 1])
        self.assertEqual(cache._refcount, {1: 1, 2: 1})
        # and after
        cache = LRUCache(maxsize=10)
        [cache.get(1) for i in range(49)]
        cache.get(2)
        [cache.get(1) for i in range(50)]
        cache.get(2) # Push us one over the limit
        self.assertEqual(list(cache._queue), [1, 2])
        self.assertEqual(cache._refcount, {1: 1, 2: 1})
        self.assertInvariants(cache)

    def test_queue_emptying(self):
        # the queue is emptied when it gets too big and we cannot compact
        from wesgi import LRUCache
        cache = LRUCache(maxsize=10)
        [cache.get(i) for i in range(100)] # fill up to maxqueue
        self.assertEqual(list(cache._queue), list(range(100))) # all elements are in the queue
        self.assertEqual(len(cache._queue), 100)
        self.assertEqual(cache._refcount, dict([(i, 1) for i in range(100)]))
        cache.get(100) #push us over the limit
        self.assertEqual(len(cache._queue), 80)
        self.assertEqual(list(cache._queue), list(range(21, 101)))
        self.assertEqual(cache._refcount, dict([(i, 1) for i in range(21, 101)]))
        self.assertInvariants(cache)

    def test_queue_emptying_memory_leak(self):
        # When we empty the queue, we need to make sure that elements in our cache stay in the queue
        from wesgi import LRUCache
        cache = LRUCache(maxsize=10)
        cache.set('x', 'y')
        [cache.get(i) for i in range(100)] # fill queue over maxsize
        self.assertEqual(cache._cache, {'x': 'y'})
        self.assertEqual(len(cache._queue), 80)
        self.assertEqual(list(cache._queue), ['x'] + list(range(21, 100)))
        expected_refcount = dict([(i, 1) for i in range(21, 100)])
        expected_refcount['x'] = 1
        self.assertEqual(cache._refcount, expected_refcount)
        self.assertInvariants(cache)

    def test_thread_fuzzing(self):
        from wesgi import LRUCache
        import threading
        import time
        max = 100
        no_threads = 2
        if all_tests:
            max = 500
            no_threads = 5
        cache = LRUCache(maxsize=5)
        def pound():
            for k in range(max):
                for i in range(10):
                    val = cache.get(i)
                    if not val:
                        cache.set(i, str(k))
                    cache.get(i - 1)
                    cache.get(i - 3)
                    cache.get(i + 3)
                    cache.delete(i + 1)
                    cache.set(i + 2, str(k))
        threads = []
        for i in range(no_threads):
            threads.append(threading.Thread(target=pound))
        for t in threads:
            t.start()
        for t in threads:
            t.join()
        self.assertInvariants(cache)

    def assertInvariants(self, cache):
        count = {}
        for k in cache._queue:
            count[k] = count.setdefault(k, 0) + 1
        self.assertEqual(count, cache._refcount)
        for k in cache._cache:
            self.assertTrue(k in count, k)

if all_tests:
    class TestRealRequest(TestCase):
        # test not run by default as it requires network connectivity

        def test_http(self):
            from wesgi import Policy
            policy = Policy()
            policy.chase_redirect = True
            http = policy.http()
            result, content = http.request('http://www.google.com/')
            self.assertTrue("google" in content.lower())

        def test_cached_http(self):
            from wesgi import LRUCache
            from wesgi import Policy
            policy = Policy()
            policy.cache = LRUCache()
            policy.chase_redirect = True
            http = policy.http()
            self.assertEqual(0, policy.cache.hits + policy.cache.misses)
            result, content = http.request('http://www.google.com/')
            self.assertTrue("google" in content.lower())
            self.assertFalse(result.fromcache)
            self.assertNotEqual(0, policy.cache.hits + policy.cache.misses)
            result, content = http.request('http://www.google.com/')
            self.assertTrue("google" in content.lower())
            self.assertNotEqual(0, policy.cache.hits + policy.cache.misses)

        def test_https(self):
            from wesgi import Policy
            policy = Policy()
            policy.chase_redirect = True
            http = policy.http()
            result, content = http.request('https://encrypted.google.com/')
            self.assertTrue("google" in content.lower())

def load_tests(loader, standard_tests, pattern):
    if all_tests:
        # run tests in our README.txt
        import doctest
        this_dir = os.path.dirname(__file__)
        readme = doctest.DocFileTest(os.path.join(this_dir, '..', 'README.rst'),
                                     module_relative=False)
        standard_tests.addTest(readme)
    return standard_tests



import os

from setuptools import setup, find_packages

_here = os.path.dirname(__file__)

_readme = os.path.join(_here, 'README.rst')
_readme = open(_readme, 'r').read()

_changes = os.path.join(_here, 'CHANGES.rst')
_changes = open(_changes, 'r').read()

setup(name='wesgi',
      version='0.10',
      description='A WSGI middleware which processes ESI directives',
      url="http://pypi.python.org/pypi/wesgi",
      long_description='%s\n%s' % (_readme, _changes),
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        ],
      keywords='web middleware wsgi esi',
      author="Brian Sutherland",
      author_email="brian@vanguardistas.net",
      packages=find_packages(),
      install_requires=["WebOb",
                        "httplib2",
                        "mock"],
      include_package_data=True,
      zip_safe=False,
      test_suite="wesgi.tests",
      )

